/*

==== String compile(String code, String filename) ====

Compiles a C program into a dynamic library named ''filename''.
Any output from the compiler is thrown as an exception.

*/

C.compile=function(code, filename) {
	var tempnam=os.tempnam()+os.dlext;
	write(tempnam,code+'\n');

// The try..finally here is commented out because it provoked a bug
// in one version of spidermonkey. Better put back in when bug is
// resolved

	try {
		C.compilefile(tempnam,filename);
	} finally {
		os.unlink(tempnam);
	}
}

/*

==== String compilefile(String filename_in, String filename_out) ====

Compiles a C program into a dynamic library named ''filename''.
Any output from the compiler is thrown as an exception.

*/

C.compilefile=function(filename_in, filename) {
	if (global._WIN32) {
		var cmd='cl /nologo /Tc '+filename_in+' '+C.wd+'\\js32.lib '+C.wd+'\\callc.lib /MD /LD /Zm400 /Fe'+filename;
		var ret=read(cmd+' |');
		var error=false;
		for each (var line in ret.split("\n")) {
			var flds=line.split(":");
			if (flds.length>1 && flds[1].indexOf('error')!=-1)
				throw(trim(ret));
		}
	} else {
		var ret=read('cc -x c '+filename_in+' --shared -o '+filename+' 2>&1 |');
		var error=false;
		for each (var line in ret.split("\n")) {
			var flds=line.split(":");
			if (flds.length>1 && flds[2].indexOf('error')!=-1)
				throw(trim(ret));
		}
	}
}

/*

====cpp(code)====

Takes code, a string.
Returns a string using libcpp.

*/

C.cpp=function(code) {
	var errormsg=[null];
	var ret=_.cpp(String(code),errormsg);
	if (errormsg[0]!=null) {
		_.free(ret);
		var error=errormsg[0].string();
		_.free(errormsg[0]);
		throw(error);
	}
	ret2=ret.string();
	_.free(ret);
	return ret2;
  
}

/*

====ctoxml(code)====

Takes code, a string.
Returns an xml object.

*/

C.ctoxml=function(code) {
	var errorpos=[-1];
	var ret=_.ctoxml(code,errorpos);
	if (errorpos[0]!=-1) {
		_.free(ret);
		throw("C.ctoxml: syntax error at "+errorpos);
	}

	var ret2=new XML(ret.string());
	_.free(ret);
	return ret2;
}

