/*
====cdb(file, [mode="r"])====

file can be a file object, a filename or a file descriptor
*/

function cdb(file, mode) {
	if (mode==undefined) mode="r";

	this.file=file;
	switch(mode) {
	case "r": this.init(); break;
	case "w": this.start(); break;
	default: throw("gdbm: Wrong mode '"+mode+"'");
	}
}

/**/

// used by open.js

cdb.open_uri=function(uri, mode) {
	return new cdb(uri.path, mode);
}

/**/

cdb.prototype.init=function() {
	switch(typeof(this.file)) {
		case "string": var fd=os.open(this.file, os.O_RDONLY); break;
		case "number": var fd=file; break;
		case "object": var fd=file.fileno(); break;
	}
	this.cdbp=_.new_struct_cdb();
	this.cdbfp=_.new_struct_cdb_find();
	_.cdb_init(this.cdbp,fd);
}

/**/

cdb.prototype.start=function() {
	switch(typeof(this.file)) {
		case "string": var fd=os.open(this.file, os.O_RDWR | os.O_CREAT | os.O_TRUNC); break;
		case "number": var fd=file; break;
		case "object": var fd=file.fileno(); break;
	}
	this.cdbmp=_.new_struct_cdb_make();
	_.cdb_make_start(this.cdbmp,fd);
}

/**/

cdb.prototype.free=function() {
	_.cdb_free(this.cdbp);
	switch(typeof(this.file)) {
		case "string": os.close(this.cdbp.cdb_fd); break;
	}
	_.free(this.cdbp);
	_.free(this.cdbfp);
}


/*
====fileno()====
*/


cdb.prototype.fileno=function() {
	if (this.cdbp!=undefined) return this.cdbp.cdb_fd;
	if (this.cdbmp!=undefined) return this.cdbmp.cdb_fd;
}


/*

====close()====

Frees any memory allocated by the cdb library. If a file name was given to [[cdb]], the file is closed.
If a file object or file descriptor was given, the file remains open.

*/

cdb.prototype.close=function() {
	if (this.cdbp!=undefined) this.free();
	if (this.cdbmp!=undefined) this.finish();
}

/**/

cdb.prototype._read=function(len, pos) {
	var buf=new Pointer(len);
	var res=_.cdb_read(this.cdbp,buf,len,pos);
	if (res==0) return buf.string(len);
}

/**/

cdb.prototype.readdata=function() {
	return this._read(this.cdbp.cdb_vlen, this.cdbp.cdb_vpos);
}

/**/

cdb.prototype.readkey=function() {
	return this._read(this.cdbp.cdb_klen, this.cdbp.cdb_kpos);
}

/**/

cdb.prototype.find=function(key) {
	var res=_.cdb_find(this.cdbp, key, key.length);
	return res?true:false;
}

/**/

cdb.prototype.findinit=function(key) {
	var res=_.cdb_findinit(this.cdbfp, this.cdbp, key, key.length);
	return res?true:false;
}

/**/

cdb.prototype.findnext=function() {
	var res=_.cdb_findnext(this.cdbfp);
	return res?true:false;
}

/**/

cdb.prototype.seqinit=function() {
	this.cptr=[2048];
}

/**/

cdb.prototype.seqnext=function() {
	if (this.cptr==undefined) this.seqinit();
	var res=_.cdb_seqnext(this.cptr,this.cdbp);
	return res?true:false;
}

/**/

cdb.prototype._exists=function(key) {
	var res=_.cdb_make_exists(this.cdbmp, key, key.length);
	return res?true:false;
}

/**/

cdb.PUT_ADD=_.CDB_PUT_ADD;
cdb.PUT_REPLACE=_.CDB_PUT_REPLACE;
cdb.PUT_INSERT=_.CDB_PUT_INSERT;
cdb.PUT_WARN=_.CDB_PUT_WARN;

cdb.prototype.put=function(key, val, flag) {
	var ret=_.cdb_make_put(this.cdbmp, key, key.length, val, val.length, flag);
	return ret?true:false;
}

/**/

cdb.prototype.add=function(key, val) {
	var res=_.cdb_make_add(this.cdbmp, key, key.length, val, val.length);
}

/**/

cdb.prototype.replace=function(key, val) {
	return this.put(key, val, cdb.PUT_REPLACE);
}

/**/

cdb.prototype.insert=function(key, val) {
	return this.put(key, val, cdb.PUT_INSERT);
}

/**/

cdb.prototype.warn=function(key, val) {
	return this.put(key, val, cdb.PUT_WARN);
}

/**/

cdb.prototype.finish=function() {
	var ret=_.cdb_make_finish(this.cdbmp);
	switch(typeof(this.file)) {
		case "string": os.close(this.cdbmp.cdb_fd); break;
	}
	_.free(this.cdbmp);
}

/*
====get(key, [unique=true])====
*/

cdb.prototype.get=function(key, unique) {
	if (unique==undefined) unique=true;
	if (unique==true) {
		if (this.find(key)) {
			return this.readdata();
		} else {
			return;
		}
	} else {
		var ret=[];
		if (this.findinit(key)) {
			do {
				ret.push(this.readdata());
			} while(this.findnext());
		}
		return ret;
	}
}

/*
====set(key, value, [replace=true])====
*/

cdb.prototype.set=function(key, value, replace) {
	if (replace==undefined) replace=true;
	if (replace==true) {
		return this.replace(key,value);
	} else {
		return this.add(key,value);
	}
}

/*
====exists(key)====
*/

cdb.prototype.exists=function(key) {
	if (this.cdbmp!=undefined) return this._exists(key);
	else return this.find(key);
}

/*
====reset()====

Reset iterator. Use with [[cdb.each]].
*/

cdb.prototype.reset=cdb.prototype.seqinit;

/*
====each()====

Return [key, value] or ''undefined'' and advance iterator
*/

cdb.prototype.each=function() {
	if (!this.seqnext()) return;
	return [this.readkey(),this.readdata()];
}


/*
====read([unique=true])====

Reads all values in database and returns an object with the same
key/value associations.

If unique is true (default), all values are strings.

If unique is false, all values are arrays of strings.
*/

cdb.prototype.read=function(unique) {
	if (unique==undefined) unique=true;

	var ret={};

	if (unique==true) {
		this.reset();
		var pair;
		while ((pair=this.each())) {
			ret[pair[0]]=pair[1];
		}
	} else {
		this.reset();
		var pair;
		while ((pair=this.each())) {
			if (pair[0] in ret) ret[pair[0]].push(pair[1]);
			else ret[pair[0]]=[pair[1]];
		}
	}
	return ret;
}


/*
====write(obj)====

Writes properties of obj to database. All properties should be strings.
*/

cdb.prototype.write=function(obj) {
	var i;
	for (i in obj) {
		this.set(i,obj[i]);
	}
}

