// contains win32-specific functions

os.sep='\\';
os.pathsep=';';
os.dlext='.dll';

/*
listdir([path="."]) -> list_of_strings
Return a list containing the names of the entries in the directory.
 
        path: path of directory to list
 
The list is in arbitrary order.  It does not include the special
entries '.' and '..' even if they are present in the directory.
*/

os.listdir=function(path) {
	if (path==undefined) path=os.curdir;
	if (path[path.length-1]==os.sep) path=path.substr(0,path.length-1);
	path+=os.sep+"*";
	var fileinfo=_.New(_.struct._finddata_t);

	var ret=[];
    var d=_._findfirst(path, fileinfo);
	if (d==-1) return ret;

	for(;;) {
		var str=String.fromCharCode.apply(String,fileinfo.name);
		var len=str.indexOf("\0");
		str=str.substr(0,len);
		if (str!="." && str!="..") ret.push(str);
		if (_._findnext(d,fileinfo)==-1) break;
	}

	_._findclose(d);
	return ret;
}

/*
stat(path) -> (mode,ino,dev=[major,minor],nlink,uid,gid,size,atime,mtime,ctime)

Perform a stat system call on the given path. Times are returned as date objects.
*/

os.stat=function(path) {
	var ret=_.New(_.struct._stat);
	if (_._stat(path, ret)==-1) {
		return null;
	}
	return os.unistat(ret);
}

/**/

os.unistat=function(buf) {
	var i;
	var j;
	var s=buf.get();
	var r={};
	for (i in s) {
		j=i.substr(3);
		switch (i) {
		case "st_atime":
		case "st_ctime":
		case "st_mtime":
			r[j]=new Date(s[i]*1000);
			break;
		default:
			r[j]=s[i];
		}
	}
	return r;
}

os.isdir=function(path) {
	var ret=_.New(_.struct._stat);
	if (_._stat(path,ret)==-1) return false;
	if (ret.st_mode & _._S_IFDIR) return true;
	return false;
}

os.isrelative=function(path) {
	return (path[0]!=os.sep && (path[1]!=':' || path[2]!=os.sep));
}

/*
errno() -> Number
Returns last error number
*/

os.errno=function() {
	return _._errno();
}

_.mkdir=function(fn, mode) {
	return _._mkdir(fn);
}

_.isatty=_._isatty;
_.chdir=_._chdir;
_.getcwd=_._getcwd;
_.stat=_._stat;
_.read=_._read;
_.close=_._close;
_.dup=_._dup;
_.open=_._open;
_.write=_._write;
_.tell=_._tell;
_.access=_._access;
_.tempnam=_._tempnam;
_.tmpnam=_._tmpnam;
_.unlink=_._unlink;
_.popen=_._popen;
_.rmdir=_._rmdir;
_.pclose=_._pclose;
_.R_OK=04;
_.W_OK=02;
_.F_OK=00;