function serializer(source) {
	this.consumeObject(source, [0], this);
}

serializer.prototype.consumeObject=function(source, atPos, obj) {
	var colonpos;
	
	while (source && source.length>atPos[0] && source[atPos[0]]!="}") {
		colonPos=source.indexOf(":",atPos[0]);
		var name=source.substr(atPos,colonPos-atPos[0]);
		atPos[0]=colonPos+1;
		obj[name]=this.consumeValue(source, atPos);
		if (source[atPos[0]]==',') atPos[0]++;
	}
}

serializer.prototype.consumeArray=function(source, atPos, obj) {
	var i=0;
	while (source && source.length>atPos[0] && source[atPos[0]]!="]") {
		obj[i]=this.consumeValue(source, atPos);
		if (source[atPos[0]]==',') atPos[0]++;
		i++;
	}
}

serializer.prototype.consumeValue=function(source, atPos) {
	switch(source[atPos[0]]) {
	case '"': // string
		atPos[0]++; // "
		var ret=this.consumeString(source, atPos);
		atPos[0]++; // "
		break;
	case '{': // string
		atPos[0]++; // {
		var ret=new Object;
		this.consumeObject(source, atPos, ret);
		atPos[0]++; // }
		break;
	case '[': // array
		atPos[0]++; // [
		var ret=[];
		this.consumeArray(source, atPos, ret);
		atPos[0]++; // ]
		break;
	default: // number
		var ret=this.consumeNumber(source, atPos);
		break;
	}

	return ret;
}

serializer.prototype.consumeNumber=function(source, atPos) {
	var pos=atPos[0];
	while(source.length>pos && source[pos]!=',' && source[pos]!=']' && source[pos]!='}') pos++;
	var ret=source.substr(atPos[0], pos-atPos[0]);
	atPos[0]=pos;
	return Number(ret);
}

serializer.prototype.consumeString=function(source, atPos) {
	var ret=[];
	for(;;) {
		var quot=source.indexOf('"',atPos[0]);
		if (quot==-1) throw("serializer: Malformed data");
		if (quot>atPos[0]+1 && source[quot-1]!='\\') break;
		ret.push(source.substr(atPos[0],quot-1-atPos[0]));
		atPos[0]=quot+1;
	}
	ret.push(source.substr(atPos[0],quot-atPos[0]));
	atPos[0]=quot;
	return ret.join('"');
}

serializer.prototype.toString=function() {
	return this.produceObject(this);
}

serializer.prototype.produceObject=function(obj) {
	var i;
	var ret=[];
	for (i in obj) {
		var v=this.produceValue(obj[i]);
		if (typeof(v)!="undefined") {
			ret.push(i+':'+v);
		}
	}
	return ret.join(",");
}

serializer.prototype.produceArray=function(obj) {
	var i;
	var ret=[];
	for (i in obj) {
		var v=this.produceValue(obj[i]);
		if (typeof(v)!="undefined") {
			ret.push(v);
		}
	}
	return ret.join(",");
}


serializer.prototype.produceValue=function(val) {
	switch(typeof(val)) {
	case "string":
		return '"'+val.replace('"','\\"')+'"';
	case "number":
		return val;
		break;
	case "object":
		if (val instanceof Array) {
			return '['+this.produceArray(val)+']';
		} else {
			return '{'+this.produceObject(val)+'}';
		}
	}
}

