function Registry(Key, pathparts, mode) {

	Key=[Key];

	if (pathparts!=undefined) {
		if (mode==undefined) mode="r";
		var sam=0;
		if (mode.indexOf("r")!=-1) sam |= _.KEY_READ;
		if (mode.indexOf("w")!=-1) sam |= _.KEY_WRITE | _.KEY_CREATE_LINK;

		for (var i=0; i<pathparts.length; i++) {
			var OldKey=Key[0];
			var res;
			var thissam=_.KEY_READ;
			if (i==pathparts.length-1) thissam=sam;
			if (res=_.RegOpenKeyEx(OldKey, pathparts[i], 0, thissam, Key)!=
				_.ERROR_SUCCESS)
				throw("Registry.OpenKey '"+pathparts[i]+"': "+windows.ErrorMessage(res));
			if (i>0) _.RegCloseKey(OldKey);
		}
	}

	this.Key=Key[0];
}

Registry.open=function(path, mode) {

	var pathparts=path.split(os.sep);

	var Key;
	var path0=pathparts.shift();
	if (path0.substr(0,5)!="HKEY_" || (Key=_[path0])==undefined)
		throw("registry.open: illegal registry path '"+path+"'");

	return new windows.Registry(Key, pathparts, mode);
}

Registry.open_uri=function(uri, mode) {
	return this.open(uri.host+uri.path.replace(/\//g,os.sep), mode);
}

Registry.prototype.read=function() {
	var Values=this.EnumValues();
	var ret=new XMLList;
	for each (var Value in Values) {
		ret+=this.QueryValue(Value);
	}
	var keys=this.EnumKeys();
	for each (var key in keys) {
		var r=new windows.Registry(this.Key, [key]);
		var ik=r.QueryInfoKey();
		ret+=<key name={key} class={ik.Class} lastWriteTime={ik.LastWriteTime}>{r.read()}</key>
		r.close();
	}
	return ret;
}

Registry.prototype.write=function(Data) {
	for (var i=0; i<Data.length(); i++) {
		var Elem=Data[i];
		switch (String(Elem.name())) {
		case "key":
			var r=new windows.Registry(this.Key);
			r.CreateKey(Elem);
			if (Elem.*.length())
				r.write(Elem.*);
			r.close();
			break;
		case "value":
			this.SetValue(Elem);
			break;
		}
	}
}

Registry.prototype.closed=false;

Registry.prototype.close=function() {
	if (!this.closed) {
		_.RegCloseKey(this.Key);
		this.closed=true;
	}
}

Registry.prototype.clear=function() {
	var subkeys=this.EnumKeys();
	for each (var key in subkeys) {
		this.DeleteKey(key, true);
	}
	var values=this.EnumValues();
	for each (var value in values) {
		this.DeleteValue(value);
	}
}

Registry.prototype.DeleteKey=function(SubKey, recursive) {
	if (recursive) {
		var r=new windows.Registry(this.Key, [SubKey], "rw");
		var subkeys=r.EnumKeys();
		for each (var key in subkeys) {
			r.DeleteKey(key, true);
		}
		r.close();
	}
	var res;
	if (res=_.RegDeleteKey(
		this.Key,
		SubKey
	)!=_.ERROR_SUCCESS)
		throw("Registry.DeleteKey '"+SubKey+"': "+windows.ErrorMessage(res));
}

Registry.prototype.DeleteValue=function(Value) {
	var res;
	if (res=_.RegDeleteValue(
		this.Key,
		Value
	)!=_.ERROR_SUCCESS)
		throw("Registry.DeleteValue '"+Value+"': "+windows.ErrorMessage(res));
}

Registry.prototype.EnumKeys=function() {
	var ik=this.QueryInfoKey();
	var SubKeys=ik.SubKeys;
	var ret=[];
	var LastWriteTime=Pointer(_.FILETIME);
	for (var i=0; i<SubKeys; i++) {
		var keybuflen=[ik.MaxSubKeyLen+1];
		var keybuf=new Pointer(keybuflen[0]*windows.CharSize);
		var res;
		if (res=_.RegEnumKeyEx(
			this.Key,
			i,
			keybuf,
			keybuflen,
			0,
			null,
			null,
			LastWriteTime)!=_.ERROR_SUCCESS)
			throw("Registry.EnumValue "+windows.ErrorMessage(res));
		ret.push(windows.Str.call(keybuf,keybuflen[0]));
	}
	return ret;
}


Registry.prototype.EnumValues=function() {
	var ik=this.QueryInfoKey();
	var nValue=ik.Values;
	var ret=[];
	for (var i=0; i<nValue; i++) {
		var valuebuflen=[ik.MaxValueNameLen+1];
		var valuebuf=new Pointer(valuebuflen[0]*windows.CharSize);
		var res;
		if (res=_.RegEnumValue(
			this.Key,
			i,
			valuebuf,
			valuebuflen,
			0,
			null,
			null,
			null)!=_.ERROR_SUCCESS)
			throw("Registry.EnumValue "+windows.ErrorMessage(res));
		ret.push(windows.Str.call(valuebuf,valuebuflen[0]));
	}
	return ret;
}


Registry.prototype.CreateKey=function(Key) {
	var newKey=[0];
	var dispos=[0];
	var res;
	if (res=_.RegCreateKeyEx(
		this.Key,
		String(Key.@name),
		0,
		String(Key.@class),
		Key.@volatile?_.REG_OPTION_VOLATILE:_.REG_OPTION_NON_VOLATILE,
		_.KEY_WRITE,
		null,
		newKey,
		dispos
	)!=_.ERROR_SUCCESS)
		throw("Registry.CreateKey '"+Key.@name+"': "+windows.ErrorMessage(res));

	this.Key=newKey[0];
}

Registry.prototype.QueryInfoKey=function() {
	var trylen=512;
	for (;;) {
		var classbuf=Pointer(trylen);
		var classbuflen=[trylen/windows.CharSize];
		var SubKeys=[0];
		var MaxSubKeyLen=[0];
		var MaxClassLen=[0];
		var Values=[0];
		var MaxValueNameLen=[0];
		var MaxValueLen=[0];
		var SecurityDescriptor=[0];
		var LastWriteTime=Pointer(_.FILETIME);
		var res;
		if (res=_.RegQueryInfoKey(
			this.Key,
			classbuf,
			classbuflen,
			0,
			SubKeys,
			MaxSubKeyLen,
			MaxClassLen,
			Values,
			MaxValueNameLen,
			MaxValueLen,
			SecurityDescriptor,
			LastWriteTime
		)!=_.ERROR_SUCCESS)
			throw("Registry.QueryInfoKey: "+windows.ErrorMessage(res));
		if (classbuflen[0]<trylen) break;
		trylen*=2;
	}
	var systime=Pointer(_.SYSTEMTIME);
	_.FileTimeToSystemTime(LastWriteTime, systime);
	systime=systime.get();

	return {
		Class: windows.Str.call(classbuf,classbuflen[0]),
		SubKeys: SubKeys[0],
		MaxSubKeyLen: MaxSubKeyLen[0],
		MaxClassLen: MaxClassLen[0],
		Values: Values[0],
		MaxValueNameLen: MaxValueNameLen[0],
		MaxValueLen: MaxValueLen[0],
		SecurityDescriptor: SecurityDescriptor[0],
		LastWriteTime: new Date(
			systime.wYear,
			systime.wMonth-1,
			systime.wDay,
			systime.wHour,
			systime.wMinute,
			systime.wSecond,
			systime.wMilliseconds)
	}
}


Registry.typenames={};

for each (var i in ["REG_BINARY", "REG_NONE", "REG_RESOURCE_LIST",
	"REG_DWORD_LITTLE_ENDIAN", "REG_DWORD_BIG_ENDIAN", "REG_EXPAND_SZ", "REG_DWORD",
	"REG_SZ", "REG_LINK", "REG_MULTI_SZ"]) 
	Registry.typenames[_[i]]=i;

Registry.prototype.SetValue=function(Value) {
	var Data;
	var Datalen;
	var x=null;
	var type=_[Value.@type];

	if (type===undefined) type=Value.@type;

	switch (type) {
	case _.REG_BINARY:
	case _.REG_NONE:
	case _.REG_RESOURCE_LIST:
		Data=String(Value.text());
		Datalen=Data.length;
		break;
	case _.REG_DWORD_LITTLE_ENDIAN:
		var num=Number(Value.text());
		Data=[
			num & 0xff,
			(num>>8) & 0xff,
			(num>>16) & 0xff,
			(num>>24) & 0xff
		];
		Datalen=4;
		break;
	case _.REG_DWORD_BIG_ENDIAN:
		var num=Number(Value.text());
		Data=[
			(num>>24) & 0xff,
			(num>>16) & 0xff,
			(num>>8) & 0xff,
			num & 0xff
		];
		Datalen=4;
		break;
	case undefined:
	case _.REG_EXPAND_SZ:
	case _.REG_SZ:
		var str=String(Value.text());
		x=_.StrDup(str);
		Data=x;
		Datalen=(str.length+1)*windows.CharSize;
		break;
	case _.REG_LINK:
		var str=String(Value.text());
		x=_.StrDupW(str);
		Data=x;
		Datalen=str.length*2;
		break;
	case _.REG_MULTI_SZ:
		var str="";
		for (var i=0; i<Value.string.length(); i++) {
			str+=Value.string[i].text()+"\0";
		}
		x=_.StrDup(str);
		Data=x;
		Datalen=(str.length+1)*windows.CharSize;
		break;
	default:
		Data=String(Value.text());
		Datalen=Data.length;
	}

	var res=_.RegSetValueEx(
		this.Key,
		String(Value.@name),
		0,
		type,
		Data,
		Datalen
	);
	if (x!=null) _.LocalFree(x);
	if (res!=_.ERROR_SUCCESS)
		throw("SetValue '"+Value.@name+"': "+windows.ErrorMessage(res));
}

Registry.prototype.QueryValue=function(ValueName) {
	var trylen=512;
	for (;;) {
		var Data=Pointer(trylen);
		var Datalen=[trylen];
		var type=[0];
		var res;
		if (res=_.RegQueryValueEx(
			this.Key,
			ValueName,
			null,
			type,
			Data,
			Datalen
		)!=_.ERROR_SUCCESS)
			throw("QueryValue '"+ValueName+"': "+windows.ErrorMessage(res));
		if (Datalen[0]<trylen) break;
		trylen*=2;
	}

	var value;
	switch (type[0]) {
	case _.REG_BINARY:
	case _.REG_NONE:
	case _.REG_RESOURCE_LIST:
		value=Data.string(Datalen[0]);
		break;
	case _.REG_DWORD_LITTLE_ENDIAN:
		var tmpstr=Data.string(4);
		value=tmpstr.charCodeAt(3)<<24 |
			tmpstr.charCodeAt(2)<<16 |
			tmpstr.charCodeAt(1)<<8 |
			tmpstr.charCodeAt(0);
		break;
	case _.REG_DWORD_BIG_ENDIAN:
		var tmpstr=Data.string(4);
		value=tmpstr.charCodeAt(0)<<24 |
			tmpstr.charCodeAt(1)<<16 |
			tmpstr.charCodeAt(2)<<8 |
			tmpstr.charCodeAt(3);
		break;
	case _.REG_EXPAND_SZ:
	case _.REG_SZ:
		value=windows.Str.call(Data);
		break;
	case _.REG_LINK:
		Data.UCstring(Datalen[0]/2);
		break;
	case _.REG_MULTI_SZ:
		value=new XMLList;
		var str=windows.Str.call(Data,Datalen[0]/windows.CharSize-2);
		str=str.split("\0");
		for each (var s in str)
			value+=<string>{s}</string>;
		break;
	default:
		value=Data.string(Datalen[0]);
	}

	var typename=windows.Registry.typenames[type[0]];
	if (typename==undefined) typename=sprintf("%08x",type[0]);

	return <value name={ValueName} type={typename}>{value}</value>;
}
