/*
* Copyright (c) 2006, Berges Allmenndigitale Rdgivningstjeneste
* All rights reserved.
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Berges Allmenndigitale Rdgivningstjeneste nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE BERGES AND CONTRIBUTORS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE BERGES AND CONTRIBUTORS BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "c_to_js2.h"
#include "pointer.h"
#include "type.h"
#include "../js/src/jsfun.h"
#include "callc.h"
#include <string.h>

size_t JSX_C_to_JS_multi(JSContext * cx, JSObject * this,
			  char *argbuf, struct JSX_type **C, int nArgs_C, jsval *v, int nArgs_v, int do_clean, int align) {
  size_t buf_size=0;

  char *C_var=argbuf;
  int i;
  for (i=0; i<nArgs_v; i++) {
    size_t elem_size;
    struct JSX_type *Ci;
    if (i<nArgs_C && do_clean && (C[i]->Const || C[i]->Dim==0)) {
      elem_size=JSX_sizeof(cx, this, C[i], v+i);
    } else {
      Ci=i<nArgs_C?C[i]:0;
      if (Ci && Ci->Dim && *Ci->Array_size) {
	// fixed array, deref the pointer
	char *fixarray=*(char **)C_var;
	JSX_C_to_JS(cx, this, fixarray, Ci, v+i, do_clean, 0);
	elem_size=sizeof(char *);
      } else {
	elem_size=JSX_C_to_JS(cx, this, C_var, Ci, v+i, do_clean, 0);
      }
    }
    if (!elem_size) return 0;
    C_var+=align*((elem_size+align-1)/align);
    buf_size+=elem_size;
  }

  if (do_clean) JS_free(cx, argbuf);
  
  return buf_size;
}


int JSX_contains_pointers(struct JSX_type *C) {
  int Dim=C->Dim;
  int *Array_size=C->Array_size;
  do {
    Dim--;
    Array_size++;
    if (Dim && *Array_size==0) return 1;
  } while (Dim);
  return 0;
}


/*

Return a new Pointer or null

 */

JSBool JSX_ptr(JSContext *cx, JSObject *obj, char *C_var, struct JSX_type *C, jsval *v) {
  if (*(void **)C_var==0) {
    *v=JSVAL_NULL;
    return JS_TRUE;
  } else {
    jsval structs;
    int size=0;
    JSBool res;
    JSObject *object=JS_NewObject(cx, &JSX_pointer_class, 0, 0);
    char *t;
    struct JSX_pointer *priv;

    if (!object) {
      JSX_ReportException(cx, "Allocation error");
      return JS_FALSE;
    }

    // Copy "struct" element
    res=JS_GetProperty(cx, obj, "struct", &structs);
    if (res && structs!=JSVAL_VOID) JS_SetProperty(cx, object, "struct", &structs);

    if (C) {
      C->Dim--;
      C->Array_size++;
      size=JSX_type_size(C);
    }
    priv=(struct JSX_pointer *)JS_malloc(cx, sizeof(struct JSX_pointer)+size);
    if (!priv) return JS_FALSE;
    //    priv->finalize=0;
    t=(char *)(priv+1);
    priv->ptr=*(void **)C_var;
    priv->callback=0;
    if (C) {
      char *tb, *tt;

      priv->type=(struct JSX_type *)t;
      memset(t,0,size);
      tb=t;
	  tt=t+size;
      JSX_type_copy(cx, C, &tb, &tt);
      C->Array_size--;
      C->Dim++;
    } else {
      priv->type=0;
    }

    JS_SetPrivate(cx, object, priv);
    *v=OBJECT_TO_JSVAL(object);
    return JS_TRUE;
  }
}


/*
  convert a struct to an object, return size of C struct
*/

size_t JSX_C_struct(JSContext *cx, JSObject *this, 
		     char *C_var, struct JSX_type *C,
		    jsval *v, JSObject *object, int do_clean ) {

  struct JSX_struct *strct;
  int i;
  char *dirvar=C_var;
  int error=0;

  strct=C->Structtype;
  if (!strct) strct=JSX_lookupstruct(cx, this, C->Struct);
  if (!strct) return 0;

  if (C && C->Dim) {
    dirvar=*(char **)C_var;
  }
  
  if (do_clean!=2) {

    if (!object) object=JS_NewObject(cx, 0, 0, 0);
    
    if (!object) {
      JSX_ReportException(cx, "Allocation error");
      return JS_FALSE;
    }
  
    *v=OBJECT_TO_JSVAL(object);
    JS_AddRoot(cx, v);
  }
  for (i=0; i<strct->nMembers; i++) {
    jsval e=JSVAL_VOID;
	JS_GetProperty(cx, object, strct->Members[i]->Name, &e);
    if (JSX_C_to_JS(cx, this, dirvar+strct->Offsets[i], strct->Members[i], &e, do_clean, 0)
		==0) error=1;
    if (do_clean!=2)
      JS_SetProperty(cx, object, strct->Members[i]->Name, &e);
  }
  if (do_clean!=2)
    JS_RemoveRoot(cx, v);

  if (C && C->Dim) {
    if (do_clean)
      free(*(char **)C_var);
	if (error) return 0;
    return sizeof(char *);
  } else {
	if (error) return 0;
    return strct->Sizeof;
  }
}


/*
  Used when C->Type is unknown
*/

size_t JSX_unknown_array_to_JS(JSContext *cx, JSObject *this,
				char *C_var,
				JSObject *array, int do_clean, char *oldptr) {

  int i;
  int arraylen;
  size_t buf_size=0;

  JS_GetArrayLength(cx, array, &arraylen);
  
  for (i=0; i<arraylen; i++) {
    jsval tmp;
    size_t elem_size;
    JS_GetElement(cx, array, i, &tmp);

    if (!JSVAL_IS_NULL(tmp) && JSVAL_IS_OBJECT(tmp) &&
	JS_IsArrayObject(cx, JSVAL_TO_OBJECT(tmp))) {
      
      // Contains pointer
      
      if (do_clean !=2 && *(void **)C_var == *(void **)oldptr) {
	
	// Array has not moved, recurse into
	
	elem_size=JSX_C_to_JS(cx, this, C_var,
			       0, &tmp, do_clean, oldptr);
	if (!elem_size) {
	  JSX_ReportException(cx, "Wrong argument type 1");
	  return 0; // Serious error - could not clean up
	}
	
      } else {
	
	// Pointer has changed, store it as pointer after cleanup
	
	if (do_clean) {
	  elem_size=JSX_C_to_JS(cx, this, oldptr,
				 0, &tmp, 2, oldptr);
	}
	
	if (JS_FALSE==JSX_ptr(cx, this, C_var, 0, &tmp)) {
	  JSX_ReportException(cx, "Alloc error - memory leakage");
	  return 0; // Serious error - could not clean up
	}
      }
      
      oldptr+=elem_size;
      
    } else {
      
      // Does not contain pointer
      
      elem_size=JSX_C_to_JS(cx, this, C_var,
			     0, &tmp, do_clean, oldptr);
      
    }
    
    C_var+=elem_size;
    buf_size+=elem_size;
    
    JS_SetElement(cx, array, i, &tmp);
    
  }

  return buf_size;
}


/* 
   Used when C->Type is known
*/

size_t JSX_known_array_to_JS(JSContext *cx, JSObject *this,
			      char *C_var, struct JSX_type *C,
			      JSObject *array, int do_clean, char *oldptr) {

  int arraylength;
  int i;
  size_t buf_size=0;
  struct JSX_type C2, *C2p;

  JS_GetArrayLength(cx, array, &arraylength);

  //  size_t buf_size=JSX_sizeof(C, 0)*arraylength;
  //  char *oldptr=C_var+buf_size;

  if (C) {
    memcpy(&C2,C, sizeof(C2));
    C2.Array_size++;
    C2.Dim--;
    C2p=&C2;
  } else {
    C2p=0;
  }
  
  if (JSX_contains_pointers(C) && oldptr) {
      
    // Contains pointers

    for (i=0; i<arraylength; i++) {
      jsval tmp;
      size_t elem_size;
      JS_GetElement(cx, array, i, &tmp);
      if (do_clean!=2 && *(void **)C_var == *(void **)oldptr) {
	
	// Pointer has not changed, recurse into
	  
	elem_size=JSX_C_to_JS(cx, this, C_var,
			       C2p,
			       &tmp, do_clean, oldptr);
	if (!elem_size) {
	  JSX_ConvertError(cx, C, "Wrong argument type 2");
	  return 0; // Serious error - could not clean up
	}
	
      } else {
	
	// Pointer has changed, store it as pointer, after cleaning up
	
	if (do_clean) {

	  if (!JSVAL_IS_NULL(tmp) && JSVAL_IS_OBJECT(tmp) &&
	      JS_IsArrayObject(cx, JSVAL_TO_OBJECT(tmp))) {

	    JSX_C_to_JS(cx, this, oldptr,
			 C2p,
			 &tmp, 2, oldptr);

	  }// else {
	  //	    JSX_ConvertError(cx, C, "Wrong argument type 3");
	  //	    return 0;
	  //	  }

	}
	
	if (JS_FALSE==JSX_ptr(cx, this, C_var, C2p, &tmp)) {
	  JSX_ConvertError(cx, C, "Wrong argument type 4");
	  return 0; // Serious error - could not clean up
	}
      }
      
      oldptr+=elem_size;
      C_var+=elem_size;
      buf_size+=elem_size;
      JS_SetElement(cx, array, i, &tmp);
    }
  } else {
      
    // Does not contain pointers
      
    for (i=0; i<arraylength; i++) {
      jsval tmp;
	  size_t elem_size;

      JS_GetElement(cx, array, i, &tmp);
	
      elem_size=JSX_C_to_JS(cx, this, C_var,
			     C2p,
			     &tmp, do_clean, oldptr);
	
      if (!elem_size) return 0;

      C_var+=elem_size;
      buf_size+=elem_size;
      JS_SetElement(cx, array, i, &tmp);
    }
      
      
  }

  return buf_size;
}



/*
  If c type is unknown, C->Type, C->Struct, C->Size, C->Array_size and C->Dim are all 0.
  If c type is 0, STRUCT, VOIDSYM, RETURN or VA, C->Size is 0.
  If js type is unknown, v is JSVAL_VOID.
  C->Array_size[0] is 0 if it is a pointer to array of unknown size. Size given by jsval.
  C->Array_size[0] is >0 if it is a constant C array
*/

size_t JSX_C_to_JS(JSContext *cx, JSObject *this, 
		    char *C_var,
		    struct JSX_type *C,
		    jsval *v,
		    int do_clean, char *oldptr) {

  JSObject *object=0;
  JSObject **arrayroot=0;
  JSObject *array;

  if (!JSVAL_IS_NULL(*v) && JSVAL_IS_OBJECT(*v)) {
    object=JSVAL_TO_OBJECT(*v);
  }

  array=object;
  if (array && JS_IsArrayObject(cx,array) &&
      (C==0 || (C->Dim && C->Array_size[0]==0))) {

    // C:  Anything but a [n] array
    // JS: Array

    if (C==0 ||
	(C->Dim==0 && C->Type==VASYM) || (C->Dim==1 && C->Array_size[0]==0 && C->Type==VOIDSYM)) {

      // C:  Unknown, VA or void *
      // JS: Array
      
      int arraylength;
      size_t buf_size = 0;
      int i;
      size_t ret;
      
      JS_GetArrayLength (cx, array, &arraylength);
      
      for (i = 0; i < arraylength; i++)
	{
	  jsval tmp;
	  size_t elem_size;

	  JS_GetElement (cx, array, i, &tmp);

	  elem_size = JSX_sizeof (cx, this, 0,
					  &tmp);
	  buf_size += elem_size;
	}
      
      ret = JSX_unknown_array_to_JS (cx, this,
					     *(char **) C_var,
					     array, do_clean,
					     *(char **) (C_var+buf_size));
      if (arraylength && do_clean) JS_free(cx, *(char **) C_var);
      
      if (!ret) return 0;
      return sizeof(char *);
    }
    
    if (C && C->Array_size[0]==0 && ((C->Dim==1 &&// C->Type!=STRUCTSYM && 
	       C->Type!=VOIDSYM && C->Type!=RETURNSYM) ||
	      C->Dim>1)) {
      
      // C:  Pointer to derefable
      // JS: Array

      int arraylength;
      size_t ret;
      
      JS_GetArrayLength (cx, array, &arraylength);
      
      if (JSX_contains_pointers(C) && !oldptr)
	{

	  // Contains pointers, this is where space was allocated

	  struct JSX_type C2, *C2p;
	  size_t elem_size;

	  if (C) {
	    memcpy(&C2,C, sizeof(C2));
	    C2.Array_size++;
	    C2.Dim--;
	    C2p=&C2;
	  } else {
	    C2p=0;
	  }
	  elem_size = JSX_sizeof (cx, this, C2p, 0);
	  oldptr=*(char **) C_var + elem_size * arraylength;
	  ret= JSX_known_array_to_JS (cx, this,
				       *(char **) C_var, C,
				       array, do_clean, oldptr);
	  
	}
      else
	{
	  
	  // Does not contain pointers, or space was allocated elsewhere
	  
	  ret= JSX_known_array_to_JS (cx, this,
				       *(char **) C_var, C,
				       array, do_clean, oldptr);
	  
	  
	}

      if (arraylength && do_clean) JS_free (cx, *(char **)C_var);

      if (!ret)
	{
	  return 0;
	}
      else
	{
	  return sizeof(void *);
	}
    }

    JSX_ConvertError(cx, C, "Wrong argument type");
    return JS_FALSE;
  }

  if (C && C->Dim==0 &&
      (C->Type==STRUCTSYM || C->Type==UNIONSYM))
    // C: Direct struct/union
    // JS: Any
    return JSX_C_struct(cx, this, C_var, C, v, object, do_clean);
  
  if (C && *v==JSVAL_VOID && C->Dim==0) {
	  int32 tmp;

    // C:  Known direct
    // JS: Unknown
    
    if (do_clean!=2) {
      switch(C->Type) {
	
      case CHARSYM:
	*v=INT_TO_JSVAL(*(char *)C_var);
	break;
	  
      case SHORTSYM:
	*v=INT_TO_JSVAL(*(int16 *)C_var);
	break;
	  
      case INTSYM:
		  tmp=*(int32 *)C_var;
		  if ((C->Signed && INT_FITS_IN_JSVAL(tmp)) ||
			  INT_FITS_IN_JSVAL(tmp<<1)) *v=INT_TO_JSVAL(tmp);
		  else {
		    jsval dtmp=C->Signed?(jsdouble)tmp:(jsdouble)(unsigned int)tmp;
			  if (!JS_NewNumberValue(cx, dtmp, v)) {
				JSX_ConvertError(cx, C, "Allocation error");
			    return 0;
			  }

		  }
	break;
	  
      case IFSYM: // Boolean
	*v=BOOLEAN_TO_JSVAL(*(int32 *)C_var);
	break;
	  
      case DOUBLESYM:
	switch(C->Size) {
	case sizeof(double):
	  if (!JS_NewNumberValue(cx, *(double *)C_var, v)) {
	    JSX_ConvertError(cx, C, "Allocation error");
	    return 0;
	  }
	  break;
	case sizeof(float):
	  if (!JS_NewNumberValue(cx, *(float *)C_var, v)) {
	    JSX_ConvertError(cx, C, "Allocation error");
	    return 0;
	  }
	  break;
	default:
	  return 0;
	}
	break;
	  
      case VASYM:
	if (!JSX_ptr(cx, this, C_var, 0, v))
	  return 0;
	break;
	  
      default:
	JSX_ConvertError(cx, C, "Wrong argument type");
	return 0;
      }
    }
    return C->Size;
  }

  if (C && C->Dim>0 && C->Array_size[0]) {
    size_t ret;

	  if (C->Dim==1 && C->Type==CHARSYM && JSVAL_IS_STRING(*v)) {
		// C: char [n]
		// JS: String

		JSString *str=JS_NewStringCopyN(cx, C_var, C->Array_size[0]);
		*v=STRING_TO_JSVAL(str);
		return sizeof(char)*C->Array_size[0];
	  }

	  if (C->Dim==1 && C->Type==SHORTSYM && JSVAL_IS_STRING(*v)) {
		// C: short [n]
		// JS: String

		JSString *str=JS_NewUCStringCopyN(cx, (short *)C_var, C->Array_size[0]);
		*v=STRING_TO_JSVAL(str);
		return sizeof(short)*C->Array_size[0];
	  }


    // C:  Known [n] array
    // JS: Any

	arrayroot=0;

    if (do_clean!=2) {
      if (!(!JSVAL_IS_NULL(*v) && JSVAL_IS_OBJECT(*v)) ||
	  !JS_IsArrayObject(cx, array=JSVAL_TO_OBJECT(*v))) {
	// Create new array
	array=JS_NewArrayObject(cx, C->Array_size[0], 0);
	
	if (!array) {
	  JSX_ConvertError(cx, C, "Allocation error");
	  return 0;
	}
	
	arrayroot=&array;
	if (!JS_AddRoot(cx,arrayroot)) {
	  JSX_ConvertError(cx, C, "Root error");
	  return 0;
	}
	
      } else {
	// Use old array
	if (!JS_SetArrayLength(cx, array, C->Array_size[0])) {
	  JSX_ConvertError(cx, C, "Allocation error");
	  return 0;
	}
      }
    }

    /*

    // If these pointers can be changed, oldptr will have been set from the outside.
      if (JSX_contains_pointers(C)) {
      
      // Contains pointers
      
      if (!oldptr) oldptr=C_var+JSX_sizeof(cx, this, C, 0);
    }
    */


    ret=JSX_known_array_to_JS(cx, this,
				      C_var, C,
				      array, do_clean,
				      oldptr);
    if (arrayroot) {
      *v=OBJECT_TO_JSVAL(array);
      JS_RemoveRoot(cx,arrayroot);
    }
    
    return ret;

  }


  if ((!C || C->Dim>0) &&
      (JSVAL_IS_VOID(*v) ||
       JSVAL_IS_NULL(*v) ||
       (object &&
	JS_InstanceOf(cx, object, &JSX_pointer_class, 0)))) {
  
    // C:  Any pointer or unknown
    // JS: Unknown, null or pointer

    if (do_clean!=2) {
      
      // JSval is undef, null or Pointer.
      
      if (!JSX_ptr(cx, this, C_var, C, v)) return 0;
    }
    return sizeof(void *);
  }

	  
  if (C && object && C->Dim==1 && *C->Array_size==0 &&
      (C->Type==STRUCTSYM || C->Type==UNIONSYM)) {
    // C: Struct *
    // JS: Object
    return JSX_C_struct(cx, this, C_var, C, v, object, do_clean);
  }
  


  if (JSVAL_IS_STRING(*v)) {
    if (C==0 ||
	(C->Type==CHARSYM && C->Dim==1) ||
	(C->Type==SHORTSYM && C->Dim==1)) {

      // C:  Unknown, char * or short *
      // JS: String

      return sizeof(char *); // do nothing
    }

    if (C && C->Type==CHARSYM && C->Dim==0) {

      // C:  char
      // JS: String

      if (do_clean!=2) {
	JSString *str=JS_NewStringCopyN(cx, C_var, 1);
	*v=STRING_TO_JSVAL(str);
      }
      return sizeof(char);
    }

    if (C && C->Type==SHORTSYM && C->Dim==0) {

      // C:  short
      // JS: String

      if (do_clean!=2) {
	JSString *str=JS_NewUCStringCopyN(cx, (jschar *)C_var, 1);
	*v=STRING_TO_JSVAL(str);
      }
      return sizeof(jschar);
    }

    if (C && C->Type==VOIDSYM && C->Dim==1) {

      // C:  void *
      // JS: String

      if (do_clean!=2) {
	if (!JSX_ptr(cx, this, C_var, C, v)) return 0;
      }
      return sizeof(void *);
    }
    JSX_ConvertError(cx, C, "Type mismatch (2)");
    return 0;
  }
    
  if (JSVAL_IS_BOOLEAN(*v)) {
    if (C==0) {

      // C:  Unknown
      // JS: Boolean

      if (do_clean!=2) {
	*v=BOOLEAN_TO_JSVAL(*(int *)C_var);
      }
      return sizeof(int);
    }

    if (C && do_clean!=2) {
      switch(C->Type) {
      case CHARSYM:
      case SHORTSYM:
      case INTSYM:
      case IFSYM:

	// C:  Char, short, int, boolean
	// JS: Boolean

	switch(C->Size) {
	case sizeof(char):
	  *v=BOOLEAN_TO_JSVAL(*(char *)C_var);
	  break;
	case sizeof(int16):
	  *v=BOOLEAN_TO_JSVAL(*(int16 *)C_var);
	  break;
	case sizeof(int32):
	  *v=BOOLEAN_TO_JSVAL(*(int32 *)C_var);
	  break;
	}
	break;
      case DOUBLESYM:

	// C:  float, double
	// JS: Boolean

	switch(C->Size) {
	case sizeof(float):
	  *v=BOOLEAN_TO_JSVAL((int)*(float *)C_var);
	  break;
	case sizeof(double):
	  *v=BOOLEAN_TO_JSVAL((int)*(double *)C_var);
	  break;
	}
      }
    }
    return C->Size;
  }

  if (C && C->Dim && C->Array_size[0]==0 &&
      JSVAL_IS_NUMBER(*v)) { 
	  // C: Any pointer
	  // JS: Int
	  if (INT_FITS_IN_JSVAL(*(int *)C_var))
		*v=INT_TO_JSVAL(*(int *)C_var);
	  else {
		jsdouble tmp=(jsdouble)*(int *)C_var;
		JS_NewNumberValue(cx, tmp, v);
	  }
	  return sizeof(int);
  }

  if (JSVAL_IS_INT(*v)) {
    if (C==0) {

      // C:  Unknown
      // JS: Int

      if (do_clean!=2) {
	*v=INT_TO_JSVAL(*(int *)C_var);
      }
      return sizeof(int);
    }

    if (C && do_clean!=2) {
      switch(C->Type) {
      case CHARSYM:
      case SHORTSYM:
      case INTSYM:
      case IFSYM:

	// C:  char, short, bool, int
	// JS: Int

	switch(C->Size) {
	case sizeof(char):
	  *v=INT_TO_JSVAL(*(char *)C_var);
	  break;
	case sizeof(int16):
	  *v=INT_TO_JSVAL(*(int16 *)C_var);
	  break;
	case sizeof(int32):
	  *v=INT_TO_JSVAL(*(int32 *)C_var);
	  break;
	}
	break;
      case DOUBLESYM:

	// C:  float, double
	// JS: Int

	switch(C->Size) {
	case sizeof(float):
	  *v=INT_TO_JSVAL((int)*(float *)C_var);
	  break;
	case sizeof(double):
	  *v=INT_TO_JSVAL((int)*(double *)C_var);
	  break;
	}
      }
    }

    return C->Size;
  }

  if (JSVAL_IS_NUMBER(*v)) {
    if (C==0 || C->Type==0) {

      // C:  unknown
      // JS: Number

      if (do_clean!=2) {
	*v=INT_TO_JSVAL(*(int *)C_var);
      }
      return sizeof(int);
    }

    if (C && do_clean!=2) {
      switch(C->Type) {
      case CHARSYM:
      case SHORTSYM:
      case INTSYM:
      case IFSYM:

	// C:  char, short, int, bool
	// JS: Number

	switch(C->Size) {
	case sizeof(char):
	  if (!JS_NewNumberValue(cx, *(char *)C_var, v)) {
	    JSX_ConvertError(cx, C, "New number error");
	    return 0;
	  }
	  break;
	case sizeof(int16):
	  if (!JS_NewNumberValue(cx, *(int16 *)C_var, v)) {
	    JSX_ConvertError(cx, C, "New number error");
	    return 0;
	  }
	  break;
	case sizeof(int32):
	  if (!JS_NewNumberValue(cx, *(int32 *)C_var, v)) {
	    JSX_ConvertError(cx, C, "New number error");
	    return 0;
	  }
	  break;
	}
	break;
      case DOUBLESYM:

	// C:  float, double
	// JS: Number

	switch(C->Size) {
	case sizeof(float):
	  if (!JS_NewNumberValue(cx, *(float *)C_var, v)) {
	    JSX_ConvertError(cx, C, "New number error");
	    return 0;
	  }
	case sizeof(double):
	  if (!JS_NewNumberValue(cx, *(double *)C_var, v)) {
	    JSX_ConvertError(cx, C, "New number error");
	    return 0;
	  }
	}
      }
      return C->Size;
    }
  }

  JSX_ConvertError(cx, C, "Type mismatch (1)");
  return 0;

}
