/*
* Copyright (c) 2006, Berges Allmenndigitale Rdgivningstjeneste
* All rights reserved.
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Berges Allmenndigitale Rdgivningstjeneste nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE BERGES AND CONTRIBUTORS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE BERGES AND CONTRIBUTORS BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define BUILD_POINTER_C

// #include "../js/src/jsfun.h"
#include "../js/src/jsobj.h"

#include <stdlib.h>

#ifdef WIN32
# include <windows.h>
# include <io.h>
# include <malloc.h>
#else
# include <dlfcn.h>
# include <alloca.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include "pointer.h"
#include "struct.h"
#include "type.h"
#include "callc.h"

static JSBool
pointer_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp);
static JSBool
pointer_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp);

JSClass JSX_pointer_class={
  "Pointer",
  JSCLASS_HAS_PRIVATE, // | JSCLASS_NEW_RESOLVE,
  JS_PropertyStub,
  JS_PropertyStub,
  pointer_getProperty,
  pointer_setProperty,
  JS_EnumerateStub,
  JS_ResolveStub,
  //(JSResolveOp) pointer_resolve,
  JS_ConvertStub,
  JSX_pointer_finalize
};

JSObject *JSX_pointer_proto;

struct JSX_callback {
  char code[5]; // "call JSX_callback_exec"
  JSContext *cx;
  JSObject *Pointer;
  JSFunction *fun;
  JSObject *object;
};

#ifdef WIN32

 void  DLLACCESS *JSX_globalsym(JSContext *cx, JSObject *dl, const char *name) {
  static dlhandle globaldl=0;
  if (globaldl==0) {
    globaldl=GetModuleHandle(NULL);
  }
  if (dl) {
	  return (void *)GetProcAddress((dlhandle)JS_GetPrivate(cx, dl),name);
  }
  else return (void *)GetProcAddress(globaldl,name);
}

#else

void DLLACCESS *JSX_globalsym(JSContext *cx, JSObject *dl, const char *name) {
  static dlhandle globaldl=0;
  if (globaldl==0) {
    globaldl=dlopen(0,RTLD_LAZY);
  }
  if (dl) return (void *)dlsym((dlhandle)JS_GetPrivate(cx, dl),name);
  else return (void *)dlsym(globaldl,name);
}

#endif

/*
  Call a function by putting arguments on stack

  Need three versions, because different mechanisms are used for
  return values that are scalar, structs and floats, respectively.
 */

#define ALIGN_TYPE int
#define CALL_ALIGN sizeof(ALIGN_TYPE)

typedef struct {int a[64];} structret;

static double vacall_cdecl_double(double (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  return (*func)();
}

static structret vacall_cdecl_struct(structret (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  return (*func)();
}

static int vacall_cdecl_int(int (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  return (*func)();
}

#ifdef _WIN32

static double vacall_stdcall_double(double (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  __asm call func
}

static structret vacall_stdcall_struct(structret (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  __asm call func
}

static int vacall_stdcall_int(int (*func)(void), ALIGN_TYPE *buf, size_t len) {
  ALIGN_TYPE *stack=alloca(len*CALL_ALIGN);
  for (; len--;)
    *(stack++)=*(buf++);
  __asm call func
}

#endif

struct JSX_pointer *JSX_pointer_settype(JSContext *cx, JSObject *this, struct JSX_pointer *old, struct JSX_type *T) {
  char *tb;
  char *tt;
  struct JSX_pointer *p;
  int size=JSX_type_size(T);
  p=(struct JSX_pointer*) JS_malloc(cx, sizeof(struct JSX_pointer)+size);
  p->ptr=old->ptr;
  p->callback=old->callback;
  p->type=(struct JSX_type *)(p+1);
  tb=(char *)p->type;
  tt=tb+size;
  JSX_type_copy(cx, T, &tb, &tt);
  p->callback->cx=cx;
  JS_free(cx, old);
  return p;
}

JSBool JS_DLL_CALLBACK JSX_pointer_call_x(JSContext *cx, JSObject *obj, struct JSX_pointer *ptr, uintN argc, jsval *argv, jsval *rval) {
  union {
    double dbl;
    int i;
    structret strct;
  } ret;

  char *argbuf=0;
  size_t arg_size=0;

  struct JSX_type **C=0;
  int nArgs=0;

  struct JSX_type *Ret=0;

  if (!ptr->ptr) { // null pointers should not be pointer objects, but null objects
    ptr->ptr=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol '%s' in pointer.call", ptr->type->Name);
      return JS_FALSE;
    }
  }

  if (ptr->type->Func) {
    if ((int)argc<ptr->type->Func->nArgs || 
	((int)argc>ptr->type->Func->nArgs &&
	 !ptr->type->Func->Elipsis)) {
      JSX_ReportException(cx, "%d arguments to pointer.call, expected %d%s for %s",
			  argc, ptr->type->Func->nArgs, ptr->type->Func->Elipsis?"+":"", ptr->type->Name);
      return JS_FALSE;
    }
    
    C=ptr->type->Func->Args;
    nArgs=ptr->type->Func->nArgs;
  }

  if (nArgs || argc) {
    arg_size=JSX_JS_to_C_multi(cx, obj, &argbuf, C, nArgs, argv, argc, 1, CALL_ALIGN);
    if (!argbuf) {
//      JSX_ReportException(cx, "Error in argument inconversion in pointer.call for %s", ptr->type->Name);
      return JS_FALSE;
    }
  }

  if (ptr->type && ptr->type->Func && ptr->type->Func->Ret) Ret=ptr->type->Func->Ret;

#ifdef _WIN32
  switch(ptr->type->Func->Callconv) {
  case STDCALLSYM:
    if (Ret && Ret->Dim==0 && (Ret->Type==DOUBLESYM || Ret->Type==STRUCTSYM)) {
      if (Ret->Type==DOUBLESYM) {
        ret.dbl=vacall_stdcall_double(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	  } else {
        ret.strct=vacall_stdcall_struct(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	  }
	} else {
      ret.i=vacall_stdcall_int(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	}
	break;
  case CDECLSYM:
#endif
  
    if (Ret && Ret->Dim==0 && (Ret->Type==DOUBLESYM || Ret->Type==STRUCTSYM)) {
      if (Ret->Type==DOUBLESYM) { 
        ret.dbl=vacall_cdecl_double(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	  } else {
        ret.strct=vacall_cdecl_struct(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	  }
	} else {
      ret.i=vacall_cdecl_int(ptr->ptr, (ALIGN_TYPE *)argbuf, arg_size/CALL_ALIGN);
	}
  
#ifdef _WIN32
	break;
  }
#endif
  if (nArgs || argc) {
    if (!JSX_C_to_JS_multi(cx, obj, argbuf, C, nArgs, argv, argc, 1, CALL_ALIGN)) {
//      JSX_ReportException(cx, "Error in argument outconversion in pointer.call for %s", ptr->type->Name);
      return JS_FALSE;
    }
  }

  if (Ret) {
    *rval=JSVAL_VOID;
    JSX_C_to_JS(cx, obj, (char *)&ret, Ret, rval, 0, 0);
  } else {
    *rval=JSVAL_ZERO; // Assume int if return type is unknown
    JSX_C_to_JS(cx, obj, (char *)&ret, 0, rval, 0, 0);
  }
  
  return JS_TRUE;
}

JSBool JS_DLL_CALLBACK JSX_pointer_call(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr=JS_GetPrivate(cx,obj);
  if (!ptr || !ptr->type || ptr->type->Type!=RETURNSYM) {
    JSX_ReportException(cx, "Uninitialised Cfunction object");
    return JS_FALSE;
  }

  return JSX_pointer_call_x(cx, obj, ptr, argc, argv, rval);
}

/* analogous to function.prototype.apply */

JSBool JSX_pointer_apply(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  // arg1 is array!

  jsval *tmpv=0;
  
  JSObject *array;
  jsint arraylen;

  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments to pointer.apply");
    return JS_FALSE;
  }

  if (JS_TRUE==JSVAL_IS_OBJECT(argv[0]) &&
      NULL!=(array=JSVAL_TO_OBJECT(argv[0])) &&
      JS_GetArrayLength(cx, array, &arraylen)) {
    int i;
    tmpv=(jsval *) JS_malloc(cx,sizeof(jsval)*(arraylen+1));
    if (!tmpv) {
      JSX_ReportException(cx, "Allocation error");
      return JS_FALSE;
    }

    for (i=0; i<arraylen; i++) {
      if (JS_GetElement(cx, array, i, tmpv+i)==JS_FALSE) {
	JSX_ReportException(cx, "Missing element #%d",i);
	goto failure;
      }
    }

    if (!JSX_pointer_call(cx, obj, arraylen, tmpv, rval)) goto failure;
    JS_free(cx, tmpv);
    return JS_TRUE;
  }
  JSX_ReportException(cx, "Expected array");
  return JS_FALSE;
    
 failure:
  if (tmpv) JS_free(cx, tmpv);
  return JS_FALSE;
}

static struct JSX_callback *alloc_callback(JSContext *cx) {
  int ptrdiff;
  char *pd;
  struct JSX_callback *ret=(struct JSX_callback *)calloc(sizeof(struct JSX_callback),1);
  ret->code[0]=(char)0xe8; // call
  ptrdiff=(char *)&JSX_callback_exec-ret->code-5;
  pd=(char *)&ptrdiff;
  ret->code[1]=pd[0];
  ret->code[2]=pd[1];
  ret->code[3]=pd[2];
  ret->code[4]=pd[3];
  return ret;
}

static void free_callback(JSContext *cx, struct JSX_callback *cb) {
  JS_RemoveRoot(cx, &cb->fun);
  JS_RemoveRoot(cx, &cb->object);
}

/*
  Create a new pointer object from a string. Used in C wrappers
 */

struct JSX_pointer *JSX_pointer_new_x_x(JSContext *cx, JSObject *object, char *s, JSObject *dl) {
  char *s2=s;
  struct JSX_pointer *ptr;
  char *tb;
  char *tt;
  int typelen;

  typelen=JSX_type_size_s(&s2);
  if (typelen==-1) goto failure;
  
  ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+typelen);
  if (!ptr) goto failure;
  
  tb=(char *)(ptr+1);
  tt=tb+typelen;
  
  ptr->callback=0;
  ptr->ptr=0;
  ptr->type=(struct JSX_type *)tb;
  JSX_type_new(cx, 0, &s, &tb, &tt);

  ptr->type->dl=dl;
  JS_AddRoot(cx, (void *)&ptr->type->dl);

  goto success;
  
 failure:
  if (ptr) JS_free(cx, ptr);
  return 0;
  
 success:
  return ptr;
}

JSObject *JSX_pointer_new_x(JSContext *cx, JSObject *object, char *s, JSObject *dl) {
  struct JSX_pointer *ptr;

  if (!object) object=JS_NewObject(cx, &JSX_pointer_class, JSX_pointer_proto, 0);

  ptr=JSX_pointer_new_x_x(cx, object, s, dl);
  if (!ptr) return 0;

  JS_SetPrivate(cx,object,ptr);
  return object;
}


JSBool JS_DLL_CALLBACK JSX_pointer_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {

  JSObject *object;
  struct JSX_pointer *ptr=0;
  JSObject *arg0=0;
  JSFunction *fun;

  if (JS_FALSE==JS_InstanceOf( cx, obj, &JSX_pointer_class, 0)) { // not called with new
    object=JS_NewObject( cx, &JSX_pointer_class, 0, 0);
    *rval=OBJECT_TO_JSVAL(object);
  } else {
    object=obj;
  }
    

  if (argc>0 && JSVAL_IS_OBJECT(argv[0]) && !JSVAL_IS_NULL(argv[0]))
    arg0=JSVAL_TO_OBJECT(argv[0]);

  // new Pointer(bytes)
  // Pointer(bytes)

  if (argc==1 && JSVAL_IS_INT(argv[0])) {
    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+JSVAL_TO_INT(argv[0]));
    if (!ptr) goto failure;

    ptr->ptr=ptr+1;
    ptr->callback=0;
    ptr->type=0;

    goto success;
  }

  // not valid: new Pointer(type, [obj, _])
  // Pointer(type, [obj, _])

  if (arg0 && JS_InstanceOf(cx, arg0, &JSX_type_class, 0) && (argc==1 || argc==2) &&
      !JS_InstanceOf( cx, obj, &JSX_pointer_class, 0)) { // not called with new

    struct JSX_type *type=JS_GetPrivate(cx, arg0);
    size_t typesize=JSX_type_size(type);
    char *tb, *tt;
    size_t Sizeof=JSX_sizeof(cx, obj, type, 0);
    jsval structs;
    JSBool res;

    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+sizeof(struct JSX_type)+typesize+Sizeof);
    if (!ptr) goto failure;

    ptr->callback=0;
    ptr->type=(struct JSX_type *)(ptr+1);
    tb=(char *)ptr->type;
    tt=tb+typesize+Sizeof;
    JSX_type_copy(cx, type, &tb, &tt);
    ptr->ptr=tb;

    res=JS_GetProperty(cx, obj, "struct", &structs);
    if (res && structs!=JSVAL_VOID) JS_SetProperty(cx, object, "struct", &structs);

    if (argc==2) {
      // Initialize pointer from given value
      size_t size=JSX_JS_to_C(cx, obj, 
			      ptr->ptr, ptr->type,
			      argv+1, 1, 0); // could clean if finalizer did it
      if (size==0) {
	JSX_ReportException(cx, "Pointer: Could not convert value");
	goto failure;
      }
    } else {
      memset(ptr->ptr, 0, Sizeof);
    }

    goto success;
  }
  
  // new Pointer(pointer)
  // Pointer(pointer)

  if (argc==1 && arg0 && JS_InstanceOf(cx, arg0, &JSX_pointer_class, 0)) {
    struct JSX_pointer *other=JS_GetPrivate(cx,arg0);
    int typesize=0;
    jsval structs;
    JSBool res;

    if (other->type) {
      typesize=JSX_type_size(other->type);
    }
    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+typesize);
    if (!ptr) goto failure;

    ptr->ptr=other->ptr;
    if (other->type) {
      char *tb=(char *)(ptr+1);
      char *tt=tb+typesize;

      ptr->type=(struct JSX_type *)(ptr+1);
      
      JSX_type_copy(cx, other->type, &tb, &tt);
    } else {
      ptr->type=0;
    }

    if (other->callback) { // Allocate new callback slot and point to same js function
      ptr->callback=alloc_callback(cx);
      if (!ptr->callback) goto failure;
      ptr->callback->fun=other->callback->fun;
      ptr->callback->Pointer=object;
      ptr->callback->cx=other->callback->cx;
	  ptr->callback->object=other->callback->object;
      JS_AddRoot(cx, &ptr->callback->fun);
      ptr->ptr=ptr->callback->code;
    } else {
      ptr->callback=0;
    }

    // Copy "struct" element
    res=JS_GetProperty(cx, arg0, "struct", &structs);
    if (res && structs!=JSVAL_VOID) JS_SetProperty(cx, object, "struct", &structs);

    goto success;
  }
  
  // new Pointer(function)
  // Pointer(function)

  
  if (argc==1 && arg0 && (fun=JS_ValueToFunction(cx,argv[0]))) {
    // Get an available callback slot
    struct JSX_callback *cb=alloc_callback(cx);
    if (!cb) goto failure;

    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer));
    if (!ptr) goto failure;

    cb->fun=fun;
    JS_AddRoot(cx, &cb->fun);

    cb->Pointer=object;
    ptr->ptr=cb->code;
    ptr->callback=cb;
    ptr->type=0;
	cb->object=obj;
    JS_AddRoot(cx, &ptr->callback->object);

    goto success;
  }

  // Pointer(string, [dl])
  // new Pointer(string, [dl])
  
  if ((argc==1 || argc==2) && JSVAL_IS_STRING(argv[0])) { // Create from a type packed in a string
    JSString *S=JSVAL_TO_STRING(argv[0]);
    char *s=JS_GetStringBytes(S);
    JSObject *dl=0;

    if (argc>1 && JSVAL_IS_OBJECT(argv[1]) && !JSVAL_IS_NULL(argv[1])) {
		dl=JSVAL_TO_OBJECT(argv[1]);
	}
      

    if (!JSX_pointer_new_x(cx, object, s, dl)) return JS_FALSE;
    return JS_TRUE;
  }

  JSX_ReportException(cx, "Unknown argument type to Pointer");

 failure:
  if (ptr) JS_free(cx, ptr);
  return JS_FALSE;

 success:
  JS_SetPrivate(cx,object,ptr);
  
  return JS_TRUE;

}

void JS_DLL_CALLBACK JSX_pointer_finalize(JSContext *cx, JSObject *obj) {
  struct JSX_pointer *p=JS_GetPrivate(cx, obj);
  if (!p) return;
  if (p->callback) free_callback(cx, p->callback);
  if (p->type && p->type->dl) JS_RemoveRoot(cx, (void *)&p->type->dl);

  JS_free(cx, p);
}

void *dummyfunc(void *x) {
	return x;
}

/*
  Indirect entry point for callback from C program
 */

int JSX_callback_exec(int retadr, int arg0) {
  static char ret=(char)0xc3;

  jsval rval=JSVAL_VOID;
  struct JSX_callback *cb=(struct JSX_callback *)((&retadr)[-1]-5);
  struct JSX_pointer *ptr=JS_GetPrivate(cb->cx,cb->Pointer);
  char *args=(char *)&arg0;

  ((&retadr)[-1])=(int)&ret;

  if (ptr && ptr->type && ptr->type->Func->nArgs) {
    jsval *tmp_argv;
    int i;
    int argsize;

    tmp_argv=(jsval *)JS_malloc(cb->cx, sizeof(jsval)*ptr->type->Func->nArgs);
    if (!tmp_argv) return 0;

    for (i=0; i<ptr->type->Func->nArgs; i++)
      tmp_argv[i]=JSVAL_VOID;

    argsize=JSX_C_to_JS_multi(cb->cx, cb->Pointer, args, ptr->type->Func->Args, ptr->type->Func->nArgs, 
		       tmp_argv, ptr->type->Func->nArgs, 0, CALL_ALIGN);

    JS_CallFunction(cb->cx,cb->object,cb->fun,ptr->type->Func->nArgs,tmp_argv,&rval);

    JSX_JS_to_C_multi(cb->cx, cb->Pointer, &args, ptr->type->Func->Args, ptr->type->Func->nArgs, 
		       tmp_argv, ptr->type->Func->nArgs, 0, CALL_ALIGN);

    JS_free(cb->cx, tmp_argv);

    if (ptr->type->Func->Callconv==STDCALLSYM && ptr->type->Func->Elipsis==0) { 
	  // This is an insane hack to ensure that a callback which was
	  // declared to go to a stdcall function works with a callbackpoint
	  // which is cdecl.
	  //
	  // The hack alters the callbackpoint's stack frame:
	  //
	  // === Before alteration ===
	  // 
	  // .
	  // .
	  // param n
	  // param n-1
	  // .
	  // .
	  // param 2
	  // param 1 (args points here)
	  // return address
	  // .
	  // .
	  // .
	  //
	  // === After alteration ===
	  //
	  // .
	  // .
	  // return address
	  // &ret
	  // .
	  // .
	  // &ret
	  // &ret (args points here)
	  // &ret
	  // .
	  // .
	  // .
	  //
	  // ret is a pointer to the static int ret, which contains the
	  // machine code instruction "ret".

          void **stackbottom=(void **)dummyfunc(&retadr);
          void **stacktop=(void **)(args+argsize)-1;

	  stacktop[0]=stackbottom[0];
	  while (stackbottom<stacktop)
		  *(stackbottom++)=&ret;

	}

  } else {
    JS_CallFunction(cb->cx,JS_GetGlobalObject(cb->cx),cb->fun,0,0,&rval);
  }

  if (ptr && ptr->type->Func->Ret) {
    int d;
    size_t elem_size=JSX_JS_to_C(cb->cx, cb->Pointer, (char *)&d, ptr->type->Func->Ret, &rval, 0, 0);
    return d;
  } else {
    int d;
    size_t elem_size;
    rval=JSVAL_ZERO;
    elem_size=JSX_JS_to_C(cb->cx, cb->Pointer, (char *)&d, 0, &rval, 0, 0);
    return d;
  }

  return 0;
}

JSBool JSX_pointer_toString(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  char *buf;
  struct JSX_pointer *ptr=JS_GetPrivate(cx,obj);
  JSString *S;
  if (!ptr) return JS_FALSE;
  if (!ptr->ptr) { // null pointers should not be pointer objects, but null objects
    ptr->ptr=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol in pointer.call");
      return JS_FALSE;
    }
  }
  buf=JS_malloc(cx,13);
  sprintf(buf,"0x%08x",ptr->ptr);
  S=JS_NewString(cx,buf,strlen(buf));
  *rval=STRING_TO_JSVAL(S);
  return JS_TRUE;
}


JSBool JSX_pointer_toNumber(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr=JS_GetPrivate(cx,obj);
  long num;
  if (!ptr) return JS_FALSE;
  if (!ptr->ptr) { // null pointers should not be pointer objects, but null objects
    ptr->ptr=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol in pointer.call");
      return JS_FALSE;
    }
  }

  num=(long) ptr->ptr;
  if (INT_FITS_IN_JSVAL(num)) *rval=INT_TO_JSVAL(num);
  else JS_NewNumberValue(cx, (double)num, rval);

  return JS_TRUE;
}


JSBool JSX_pointer_add(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  int32 elems=1;
  struct JSX_pointer *ptr;
  char *P;
  int Sizeof=1;

  if (argc>1) return JS_FALSE;

  JS_ConvertArguments(cx, argc, argv, "/i", &elems);

  ptr=JS_GetPrivate(cx, obj);
  if (ptr==0) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }
  P=(char *)ptr->ptr;
  if (!P) {
    if (ptr->type) ptr->ptr=P=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);

    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol");
      return JS_FALSE;
    }
  }

  if (ptr->type) {
    Sizeof=JSX_sizeof(cx, obj, ptr->type, 0);
  }

  P+=elems*Sizeof;
  ptr->ptr=P;

  return JS_TRUE;
}

JSBool JSX_pointer_sizeof(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr;
  struct JSX_type *type=0;
  int Sizeof=1;

  if (argc==1) {
    // Argument 1 should be a type or pointer
    JSObject *object=0;

    if (JSVAL_IS_OBJECT(argv[0])) 
      object=JSVAL_TO_OBJECT(argv[0]);

    if (object && JS_InstanceOf( cx, object, &JSX_type_class, 0)) {
      type=JS_GetPrivate(cx, object);
    } else if (object && JS_InstanceOf( cx, object, &JSX_pointer_class, 0)) {
      ptr=JS_GetPrivate(cx, object);
      type=ptr->type;
    } else {
      JSX_ReportException(cx, "Invalid argument to sizeof");
      return JS_FALSE;
    }
  } else {
    ptr=JS_GetPrivate(cx, obj);
    type=ptr->type;
  }

  if (type==0) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }

  if (type) {
    Sizeof=JSX_sizeof(cx, obj, type, 0);
  }

  *rval=INT_TO_JSVAL(Sizeof);
  return JS_TRUE;
}

void *JSX_pointer_member_x(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, struct JSX_type **T, int report) {
  JSString *S;
  char *s;
  struct JSX_pointer *ptr=(struct JSX_pointer *)JS_GetPrivate(cx,obj);
  if (!ptr) {
    *T=0;
    return (void *)JSX_pointer_member_x;
  }

  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments to pointer.member");
    return 0;
  }

  // Named member of struct or union

  if (JSVAL_IS_STRING(argv[0]) &&
      (S=JSVAL_TO_STRING(argv[0])) &&

      (s=JS_GetStringBytes(S))) {

    struct JSX_struct *strct;
    int membernumber;
    struct JSX_type *C=ptr->type;

    if (!C || C->Dim ||
	(C->Type != STRUCTSYM && C->Type != UNIONSYM)) {
      if (report)
	JSX_ReportException(cx, "Wrong pointer type in pointer.member");
      return 0;
    }
    strct=C->Structtype;
    if (!strct) strct=JSX_lookupstruct(cx, obj, C->Struct);
    if (!strct) {
      if (report)
	JSX_ReportException(cx, "Unknown struct %s", C->Struct);
      return 0;
    }
    
    membernumber=JSX_struct_member(strct,s);
    
    if (membernumber==-1) {
      *T=0;
      return (void *)JSX_pointer_member_x;
    }

    *T=strct->Members[membernumber];
    return (void *)((char *)ptr->ptr+strct->Offsets[membernumber]);
    
  }

  // Numbered member of array

  if (JSVAL_IS_INT(argv[0])) {

    int Sizeof;
    int membernumber=JSVAL_TO_INT(argv[0]);

    struct JSX_type *C=ptr->type;
    if (!C) {
      if (report)
	JSX_ReportException(cx, "Wrong pointer type in pointer.member");
      return JS_FALSE;
    }

    *T=C;
    if (C->Dim && *C->Array_size) {
      C->Dim--;
      C->Array_size++;
      Sizeof=JSX_sizeof(cx, obj, C, 0);
      C->Dim++;
      C->Array_size--;
    } else {
      Sizeof=JSX_sizeof(cx, obj, C, 0);
    }

    return (void *)((char *)ptr->ptr+Sizeof*membernumber);
  }

  if (report)
    JSX_ReportException(cx, "Wrong argument type to pointer.member");
  return 0;
}

JSBool JSX_pointer_member(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_type *C;
  JSObject *object;
  jsval structs;
  JSBool res;
  char *t;
  char *tb, *tt;
  int old_array_size;
  int size;
  struct JSX_pointer *priv;

  void *p=JSX_pointer_member_x(cx, obj, argc, argv, &C, 1);

  if (!p) return JS_FALSE;
  if (C==0) {
    *rval=JSVAL_VOID;
    return JS_TRUE;
  }
  
    
  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments to pointer.member");
    return JS_FALSE;
  }

  object=JS_NewObject(cx, &JSX_pointer_class, 0, 0);
  if (!object) {
    JSX_ReportException(cx, "Allocation error");
    return JS_FALSE;
  }
  *rval=OBJECT_TO_JSVAL(object);
  
  // Copy "struct" element
  res=JS_GetProperty(cx, obj, "struct", &structs);
  if (res && structs!=JSVAL_VOID) JS_SetProperty(cx, object, "struct", &structs);
  
  if (JSVAL_IS_INT(argv[0]) && C->Dim && *C->Array_size) {
    C->Dim--;
    C->Array_size++;
    old_array_size=1;
  } else
    old_array_size=0;

  size=JSX_type_size(C);
  
  priv=(struct JSX_pointer *)JS_malloc(cx, sizeof(struct JSX_pointer)+size);
  if (!priv) return JS_FALSE;
  //    priv->finalize=0;
  t=(char *)(priv+1);
  priv->ptr=p;
  priv->callback=0;
  priv->type=(struct JSX_type *)t;
  memset(t,0,size);
  tb=t;
  tt=t+size;
  JSX_type_copy(cx, C, &tb, &tt);
  
  if (old_array_size) {
    C->Dim++;
    C->Array_size--;
  }

  JS_SetPrivate(cx, object, priv);
  return JS_TRUE;

}

static JSBool
pointer_setProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp) {
  struct JSX_type *C;
  void *p;
  size_t s;
  int old_array_size;
  JSString *idstr;
  char *name;

  idstr = JS_ValueToString(cx, id);
  name = JS_GetStringBytes(idstr);

  if (strcmp(name,"struct")==0) return JS_TRUE;

  p=JSX_pointer_member_x(cx, obj, 1, &id, &C, 0);
  if (p==0 || C==0) {
    return JS_TRUE;
  }

  if (JSVAL_IS_INT(id) && C->Dim && *C->Array_size) {
    C->Dim--;
    C->Array_size++;
    old_array_size=1;
  } else
    old_array_size=0;

  s=JSX_JS_to_C(cx, obj, (char *)p, C, vp, 0, 0);

  if (old_array_size) {
    C->Dim++;
    C->Array_size--;
  }

  if (s==0) return JS_FALSE;
  return JS_TRUE;
  
}

static JSBool
pointer_getProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp) {

  struct JSX_type *C;
  void *p;
  char *name;
  int old_array_size;
  JSString *idstr;
  size_t s;

  idstr = JS_ValueToString(cx, id);
  name = JS_GetStringBytes(idstr);

  if (strcmp(name,"struct")==0) return JS_TRUE;

  p=JSX_pointer_member_x(cx, obj, 1, &id, &C, 0);
  if (p==0 || C==0) {
    return JS_TRUE;
  }

  if (JSVAL_IS_INT(id) && C->Dim && *C->Array_size) {
    C->Dim--;
    C->Array_size++;
    old_array_size=1;
  } else
    old_array_size=0;

  *vp=JSVAL_VOID;
  s=JSX_C_to_JS(cx, obj, (char *)p, C, vp, 0, 0);

  if (old_array_size) {
    C->Dim++;
    C->Array_size--;
  }

  if (s==0) return JS_FALSE;
  return JS_TRUE;
  
}

JSBool JSX_pointer_get(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr=(struct JSX_pointer *)JS_GetPrivate(cx,obj);
  if (!ptr->ptr) { // null pointers should not be pointer objects, but null objects
    ptr->ptr=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol in pointer.get");
      return JS_FALSE;
    }
  }
  
  if (argc!=0) {
    JSX_ReportException(cx, "Wrong number of arguments to pointer.get");
    return JS_FALSE;
  }
  if (!ptr->type) {
    JSX_ReportException(cx, "Unknown type in pointer.get");
    return JS_FALSE;
  }

  *rval=JSVAL_VOID;
  //  ptr->type->Dim--;
  //  ptr->type->Array_size++;
  JSX_C_to_JS(cx, obj, (char *)(ptr->ptr), ptr->type, rval, 0, 0);
  //  ptr->type->Dim++;
  //  ptr->type->Array_size--;

  return JS_TRUE;
}


JSBool JSX_pointer_set(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr=(struct JSX_pointer *)JS_GetPrivate(cx,obj);
  if (!ptr->ptr) { // null pointers should not be pointer objects, but null objects
    ptr->ptr=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol in pointer.set");
      return JS_FALSE;
    }
  }

  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments to pointer.set");
    return JS_FALSE;
  }

  JSX_JS_to_C(cx, obj, (char *)ptr->ptr, ptr->type, &argv[0], 0, 0);

  return JS_TRUE;
}


JSBool JSX_pointer_type(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr=0;

  if (argc==0) { // Return type packed as a astring
    int typestrlen;
    char *stringchars;
    struct JSX_pointer *ptr=(struct JSX_pointer *)JS_GetPrivate(cx, obj);
    JSString *S;

    if (!ptr || !ptr->type) {
      JSX_ReportException(cx, "Unknown type");
      return JS_FALSE;
    }
    typestrlen=JSX_type_string_len(cx, ptr->type);
    stringchars=(char *)JS_malloc(cx, typestrlen);
    if (!stringchars) {
      JSX_ReportException(cx, "Out of memory");
      return JS_FALSE;
    }
    JSX_type_string(cx, ptr->type, stringchars);
    S=JS_NewString(cx,stringchars,typestrlen);
    *rval=STRING_TO_JSVAL(S);
    return JS_TRUE;
  }

  if (argc==1 && JSVAL_IS_STRING(argv[0])) {    // Create from a type packed in a string
    char *s;
    char *s2;
    char *tb;
    char *tt;
    int typelen;
    struct JSX_pointer *old;

    JSString *S=JSVAL_TO_STRING(argv[0]);
    s=JS_GetStringBytes(S);
    s2=s;
    typelen=JSX_type_size_s(&s2);
    if (typelen==-1) goto failure;
    
    old=(struct JSX_pointer *)JS_GetPrivate(cx, obj);

    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+typelen);
    if (!ptr) goto failure;

    ptr->ptr=old->ptr;
    ptr->callback=old->callback;

    tb=(char *)(ptr+1);
    tt=tb+typelen;
    ptr->type=(struct JSX_type *)tb;
    JSX_type_new(cx, obj, &s, &tb, &tt);
    JS_free(cx,(void *)old);
    goto success;
  }

 failure:
  if (ptr) JS_free(cx, ptr);
  return JS_FALSE;

 success:
  JS_SetPrivate(cx,obj,ptr);
  
  return JS_TRUE;
}

JSBool JSX_pointer_string(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr;
  char *P;
  JSString *S;

  if (argc>1) return JS_FALSE;

  ptr=JS_GetPrivate(cx, obj);
  if (ptr==0) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }
  P=(char *)ptr->ptr;
  if (!P) {
    if (ptr->type) ptr->ptr=P=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol");
    }
  }
  
  if (argc) {
    if (!JSVAL_IS_INT(argv[0])) {
      JSX_ReportException(cx, "Unknown error");
      return JS_FALSE;
    }
    S=JS_NewStringCopyN(cx,P,JSVAL_TO_INT(argv[0]));
  }
  else S=JS_NewStringCopyZ(cx,P);

  if (!S) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }

  *rval=STRING_TO_JSVAL(S);
  return JS_TRUE;
}

JSBool JSX_pointer_UCString(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  struct JSX_pointer *ptr;
  jschar *P;
  JSString *S;

  if (argc>1) return JS_FALSE;

  ptr=JS_GetPrivate(cx, obj);
  if (ptr==0) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }
  P=(jschar *)ptr->ptr;
  if (!P) {
    if (ptr->type) ptr->ptr=P=JSX_globalsym(cx, ptr->type->dl, ptr->type->Name);
    if (!ptr->ptr) {
      JSX_ReportException(cx, "Undefined symbol");
    }
  }
  
  if (argc) {
    if (!JSVAL_IS_INT(argv[0])) {
      JSX_ReportException(cx, "Unknown error");
      return JS_FALSE;
    }
    S=JS_NewUCStringCopyN(cx,P,JSVAL_TO_INT(argv[0]));
  }
  else S=JS_NewUCStringCopyZ(cx,P);

  if (!S) {
    JSX_ReportException(cx, "Unknown error");
    return JS_FALSE;
  }

  *rval=STRING_TO_JSVAL(S);
  return JS_TRUE;
}


JSContext *JSEXT_InitContext(JSRuntime *rt);

static JSBool JS_DLL_CALLBACK JSX_InitContext(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  JSObject *object;
  JSRuntime *rt;

  JSContext *ncx;
  struct JSX_pointer *rtpriv;
  struct JSX_pointer *priv;


  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments");
    return JS_FALSE;
  }

  if (!JSVAL_IS_OBJECT(argv[0]) || JSVAL_IS_NULL(argv[0]) ||
      !JS_InstanceOf( cx, (object=JSVAL_TO_OBJECT(argv[0])), &JSX_pointer_class, 0)) {
    JSX_ReportException(cx, "Wrong argument type");
    return JS_FALSE;
  }
  
  rtpriv=JS_GetPrivate(cx,object);
  rt=rtpriv->ptr;

  ncx=JSEXT_InitContext(rt);

  obj=JS_NewObject( cx, &JSX_pointer_class, 0, object);
  if (!obj) return JS_FALSE;
  priv=(struct JSX_pointer *)JS_malloc(cx, sizeof(struct JSX_pointer));
  if (!priv) {
    JS_DestroyContext(ncx);
    return JS_FALSE;
  }
  //  priv->finalize=0;
  priv->ptr=ncx;
  JS_SetPrivate(cx, obj, priv);
  *rval=OBJECT_TO_JSVAL(obj);
  return JS_TRUE;
}

static JSBool JS_DLL_CALLBACK JSX_GC(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  JS_GC(cx);
  return JS_TRUE;
}

#if 0
static JSBool JS_DLL_CALLBACK JSX_load(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  char *filename;
  int fd;
  struct stat S;
  char *buf;
  int shebanglen=0;
  JSScript *script;

  if (argc!=1) {
    JSX_ReportException(cx, "Wrong number of arguments");
    return JS_FALSE;
  }

  if (!JS_ConvertArguments(cx, argc, argv, "s", &filename)) {
    return JS_FALSE;
  }

#ifdef WIN32
  fd=_open(filename,O_RDONLY);
#else
  fd=open(filename,O_RDONLY);
#endif
  if (fd==-1) {
    // File does not exist, throw exception
    JSX_ReportException(cx, "load: Could not open %s",filename);
    return JS_FALSE;
  }
  fstat(fd, &S);
  buf=malloc(S.st_size);
  if (!buf) {
    JSX_ReportException(cx, "load: Out of memory for %s", filename);
    goto failure;
  }
  if (
#ifdef WIN32
	  _read(fd, buf, S.st_size)
#else
	  read(fd, buf, S.st_size)
#endif
	  !=S.st_size) {
    JSX_ReportException(cx, "load: Error loading %s",filename);
    goto failure;
  }

  
  if (S.st_size>1 && buf[0]=='#' && buf[1]=='!') {
    // shebang
    while (buf[shebanglen]!='\n' &&
	   buf[shebanglen]!='\r') shebanglen++;
    shebanglen++;
    if (shebanglen<S.st_size &&
	(buf[shebanglen]=='\n' ||
	 buf[shebanglen]=='\r') &&
	buf[shebanglen]!=buf[shebanglen-1]) shebanglen++;
  }

  script=JS_CompileScript(cx, obj, buf+shebanglen, S.st_size-shebanglen, filename, shebanglen?2:1);
  if (!script) goto failure;
  
  free(buf);
  buf=0;
  
  *rval=JSVAL_ZERO;

#ifdef WIN32
  _close(fd);
#else
  close(fd);
#endif

  fd=0;
  
  if (!JS_ExecuteScript(cx, obj, script, rval)) goto failure;
  return JS_TRUE;

 failure:
  if (buf) free(buf);
#ifdef WIN32
  if (fd) _close(fd);
#else
  if (fd) close(fd);
#endif
  return JS_FALSE;
}
#endif
