#ifndef TYPE_H
#define TYPE_H

#include "../js/src/jsapi.h"

#ifdef WIN32
# include <windows.h>
  typedef HMODULE dlhandle;
#else
  typedef void *dlhandle;
#endif

enum callconv { CDECLSYM, STDCALLSYM };

struct JSX_functype {
  int nArgs;
  struct JSX_type **Args;
  struct JSX_type *Ret;
  enum callconv Callconv;
  int Elipsis;
};

// Resolve symbols that have been loaded due to
// too eager symbol resolution by dlopen pertaining to global variables

struct JSX_struct *JSX_lookupstruct(JSContext *cx, JSObject *this, char *Structname);

int JSX_functype_string_len(JSContext *cx, struct JSX_functype *Func);
int JSX_struct_string_len(JSContext *cx, struct JSX_struct *strct);
int JSX_type_string_len(JSContext *cx, struct JSX_type *type);
char *JSX_functype_string(JSContext *cx, struct JSX_functype *Func, char *t);
char *JSX_struct_string(JSContext *cx, struct JSX_struct *strct, char *t);
char *JSX_type_string(JSContext *cx, struct JSX_type *type, char *t);

void JSX_functype_copy(JSContext *cx, struct JSX_functype *F, char **tb, char **tt);
void JSX_functype_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt);
int JSX_functype_size_s(char **s);
int JSX_functype_size(struct JSX_functype *F);

struct JSX_type {
  int Type;
  size_t Size;
  int *Array_size;
  int Dim;
  char *Struct; // Refers to a property of the pointer object
  struct JSX_functype *Func;
  JSObject *dl;
  char *Name;
  struct JSX_struct *Structtype;
  int Signed;
  int Const;
};

void JSX_type_copy(JSContext *cx, struct JSX_type *C, char **tb, char **tt);
void JSX_type_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt);
int JSX_type_size_s(char **s);
int JSX_type_size(struct JSX_type *C);

struct JSX_struct {
  int nMembers;
  struct JSX_type **Members;
  int *Offsets;
  int Sizeof;
};

void JSX_struct_copy(JSContext *cx, struct JSX_struct *S, char **tb, char **tt);
void JSX_struct_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt);
int JSX_struct_member(struct JSX_struct *S, char *name);
int JSX_struct_size_s(char **s);
int JSX_struct_size(struct JSX_struct *S);

size_t JSX_sizeof(JSContext * cx, JSObject * this, struct JSX_type *C, jsval *v);
size_t JSX_align(JSContext * cx, JSObject * this, struct JSX_type *C, jsval *v);

extern JSClass JSX_type_class;
extern JSObject *JSX_type_proto;
JSBool JS_DLL_CALLBACK JSX_Type_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);

#endif
